class TaskPred:

    obj_list = []

    def __init__(self, params):
        self.task_pred_id = params.get('task_pred_id').strip() if params.get('task_pred_id') else None
        self.task_id = int(params.get('task_id')) if params.get('task_id') else None
        self.pred_task_id = int(params.get('pred_task_id')) if params.get('pred_task_id') else None
        self.proj_id = params.get('proj_id').strip() if params.get('proj_id') else None
        self.pred_proj_id = params.get('proj_id').strip() if params.get('pred_proj_id') else None
        self.pred_type = params.get('pred_type').strip() if params.get('pred_type') else None
        self.lag_hr_cnt = float(params.get('lag_hr_cnt').strip()) if params.get('lag_hr_cnt') else None
        self.float_path = params.get('float_path').strip() if params.get('float_path') else None
        self.aref = params.get('aref').strip() if params.get('aref') else None
        self.arls = params.get('arls').strip() if params.get('arls') else None
        TaskPred.obj_list.append(self)

    def get_id(self):
        return self.task_pred_id

    def __repr__(self):
        return str(self.task_id) + '- ' + self.pred_type + ' ->' + str(self.pred_task_id) + ' lag: ' + str(self.lag_hr_cnt)


'''
This file starts the process of reading and parsing xer files

'''
import csv
import mmap
from tqdm import tqdm
import codecs
from xerparser import *


class Reader:

    """
    Reader class is the main parser for xer files. It takes a file path as a parameter.
    The read records are parsed into classes depending on their types like activities, wbs,etc.

    :param <file>: is an xer file that need to be parsed
    :return: Reader object that contains collection of parsed records
    :Example:

    from xerparser.reader import Reader
    file = Reader('file.xer')

    """

    current_table = ''
    current_headers = []

    def create_object(self, object_type, params):
        """

        Args:
            object_type:
            params:

        Returns:

        """
        if object_type.strip() == "CURRTYPE":
            self._currencies.add(params)
        elif object_type.strip() == "ROLES":
            self._roles.add(params)
        elif object_type.strip() == "ACCOUNT":
            self._accounts.add(params)
        elif object_type.strip() == "ROLERATE":
            self._rolerates.add(params)
        elif object_type.strip() == "OBS":
            self._obss.add(params)
        elif object_type.strip() == "RCATTYPE":
            self._rcattypes.add(params)
        elif object_type.strip() == "UDFTYPE":
            self._udftypes.add(params)
        elif object_type.strip() == "RCATVAL":
            self._rcatvals.add(params)
        elif object_type.strip() == "PROJECT":
            self._projects.add(params)
        elif object_type.strip() == "CALENDAR":
            self._calendars.add(params)
        elif object_type.strip() == "SCHEDOPTIONS":
            self._schedoptions.add(params)
        elif object_type.strip() == "PROJWBS":
            self._wbss.add(params)
        elif object_type.strip() == "RSRC":
            self._resources.add(params)
        elif object_type.strip() == "RSRCCURV":
            self._rsrcurves.add(params)
        elif object_type.strip() == "ACTVTYPE":
            self._acttypes.add(params)
        elif object_type.strip() == "RSRCRATE":
            self._rsrcrates.add(params)
        elif object_type.strip() == "RSRCRCAT":
            self._rsrccats.add(params)
        elif object_type.strip() == "TASK":
            self._tasks.add(params)
        elif object_type.strip() == "ACTVCODE":
            self._activitycodes.add(params)
        elif object_type.strip() == "TASKPRED":
            self._predecessors.add(params)
        elif object_type.strip() == "TASKRSRC":
            self._activityresources.add(params)
        elif object_type.strip() == "TASKACTV":
            self._activitycodes.add(params)
        elif object_type.strip() == "UDFVALUE":
            self._udfvalues.add(params)

    def summary(self):
        print('Number of activities: ', self.tasks.count)
        print('Number of relationships: ', len(TaskPred.obj_list))

    @property
    def projects(self):
        """
        Projects

        :return: list of all projects contained in the xer file

        todo:: text

        """
        return self._projects

    @property
    def activities(self):
        """
        Property to retrieve list of tasks
        Returns: list of tasks
        """
        return self._tasks

    @property
    def wbss(self):
        """
        Property to return all wbs elements in the parsed file
        Returns:
            list of wbs elements
        """
        return self._wbss

    @property
    def relations(self):
        """
        Property to retrieve relationships from parsed file
        Returns:
            list of relations
        """
        return self._predecessors

    @property
    def resources(self):
        """
        Property to return a list of resources from parsed file
        Returns:
            list of resources of type Resources
        """
        return self._resources

    @property
    def accounts(self):
        """
        Property to return a list of accounts from parsed files
        Returns:
            list of accounts of type Accounts
        """
        return self._accounts

    @property
    def activitycodes(self):
        """
        Property to return a list of activity codes from parsed files
        Returns:
            list of accounts of type ActivityCode

        """
        return self._activitycodes

    @property
    def acttypes(self):
        return self._acttypes

    @property
    def calendars(self):
        return self._calendars

    @property
    def currencies(self):
        return self._currencies

    @property
    def obss(self):
        return self._obss
    @property
    def rcattypes(self):
        return self.rcattypes

    @property
    def rcatvals(self):
        return self._rcatvals

    @property
    def rolerates(self):
        return self._rolerates

    @property
    def roles(self):
        return self._roles

    @property
    def resourcecurves(self):
        return self._rsrcurves

    @property
    def resourcerates(self):
        return self._rsrcrates

    @property
    def resourcecategories(self):
        return self._rsrccats

    @property
    def scheduleoptions(self):
        return self._schedoptions

    @property
    def activityresources(self):
        return self._activityresources

    @property
    def udfvalues(self):
        return self._udfvalues

    @property
    def udftypes(self):
        return self._udftypes

    def __init__(self, filename):
        file = open(filename, 'r')
        self._tasks = Tasks()
        self._predecessors = Predecessors()
        self._projects = Projects()
        self._wbss = WBSs()
        self._resources = Resources()
        self._accounts = Accounts()
        self._activitycodes = ActivityCodes()
        self._acttypes = ActTypes()
        self._calendars = Calendars()
        self._currencies = Currencies()
        self._obss = OBSs()
        self._rcatvals = RCatVals()
        self._rcattypes = RCatTypes()
        self._rolerates = RoleRates()
        self._roles = Roles()
        self._rsrcurves = ResourceCurves()
        self._rsrcrates = ResourceRates()
        self._rsrccats = ResourceCategories()
        self._schedoptions = SchedOptions()
        self._activityresources = ActivityResources()
        self._udftypes = UDFTypes()
        self._udfvalues = UDFValues()
        with codecs.open(filename, encoding='utf-8', errors='ignore') as tsvfile:
            stream = csv.reader(tsvfile, delimiter='\t')
            for row in tqdm(stream, total=self.get_num_lines(filename)):
                if row[0] =="%T":
                    current_table = row[1]
                elif row[0] == "%F":
                    current_headers = [r.strip() for r in row[1:]]
                elif row[0] == "%R":
                    zipped_record = dict(zip(current_headers, row[1:]))
                    self.create_object(current_table, zipped_record)

        # for line in content:
        #     line_lst = line.split('\t')
        #     if line_lst[0] == "%T":
        #         current_table = line_lst[1]
        #     elif line_lst[0] == "%R":
        #         self.create_object(current_table, line_lst[1:])

    def get_num_lines(self, file_path):
        fp = open(file_path, "r+")
        buf = mmap.mmap(fp.fileno(), 0)
        lines = 0
        while buf.readline():
            lines += 1
        return lines



import re
import datetime


class CalendarData:
    working_days = dict()
    exceptions = []

    def __init__(self, text):
        self.text = text
        cal2 = []
        cal = re.findall("\(\d\|\|\D+\(\)", text)
        for c in cal:
            c2 = c.split("||")[1]
            cal2.append(c2.split("()")[0])
        val = re.split("\(\d\|\|\D+\(\)", text.strip())
        x = 0

        self.data = dict()
        i = 0
        for c in range(len(cal2)):
            if c >= 1:
                self.data[cal2[c]] = val[c + 1]

            else:
                self.data[cal2[c]] = ''
        self.exceptions = self.get_exceptions()
        self.working_days = self.get_days()

    def xldate_to_datetime(self, xldate):
        temp = datetime.datetime(1899, 12, 30)
        delta = datetime.timedelta(days=xldate)
        return temp + delta

    def get_exceptions(self):
        if self.data.get('Exceptions'):
            exceptions = self.data['Exceptions'] if self.data['Exceptions'] else None
            except1 = re.split("\)\)\x7f\x7f", exceptions)
            clean_exceptions = []
            for e in except1:
                if e:
                    if (len(e.split("||"))) > 1:
                        clean_exceptions.append(self.xldate_to_datetime(int(re.findall("\d+", e.split("||")[1])[1])))
                    else:
                        clean_exceptions = None

            self.exceptions = clean_exceptions
        return self.exceptions

    def get_days(self):
        if not self.data.get('DaysOfWeek'):
            return
        first = re.findall("\(\d\|\|\d\(\)(.*?)\)\)\x7f\x7f", self.data.get('DaysOfWeek'))
        days = dict()
        i = 0
        for x in first:
            second = x.split("\x7f\x7f")
            key = str(i)
            if len(second) > 1:
                days[key] = (len(second[1].strip()) > 0)
            else:
                days[key] = False
            i += 1
        self.working_days = days
        return self.working_days


class UDFValue:

    udf_code_id = None
    udf_type_id = None
    fk_id = None
    proj_id = None
    udf_number = None
    udf_text = None

    def __init__(self, params):

        self.udf_type_id = params.get('udf_type_id').strip() if params.get('udf_type_id') else None
        self.fk_id = params.get('fk_id').strip() if params.get('fk_id') else None
        self.proj_id = params.get('proj_id').strip() if params.get('proj_id') else None
        self.udf_date = params.get('udf_date').strip() if params.get('udf_date') else None
        self.udf_number = params.get('udf_number').strip() if params.get('udf_number') else None
        self.udf_text = params.get('udf_text').strip() if params.get('udf_text') else None
        self.udf_code_id = params.get('udf_code_id').strip() if params.get('udf_code_id') else None

    def get_id(self):
        return self.udf_type_id

    @staticmethod
    def find_by_id(code_id, activity_code_dict):
        return {k: v for k, v in activity_code_dict.items() if v.actv_code_id == code_id}

    def __repr__(self):
        return self.udf_text + '->' + self.udf_code_id


from xerparser.model.classes.rcatval import RCatVal

class RCatVals:

    _rcatvals = []

    def __init__(self):
        self.index = 0

    def add(self, params):
        self._rcatvals.append(RCatVal(params))

    @classmethod
    def find_by_id(cls, id):
        obj = list(filter(lambda x: x.actv_code_type_id == id, cls._rcatvals))
        if len(obj) > 0:
            return obj[0]
        return obj

    @property
    def count(self):
        return len(self._rcatvals)

    def __len__(self):
        return len(RCatVals._rcatvals)

    def __iter__(self):
        return self

    def __next__(self):
        if self.index >= len(self._rcatvals):
            raise StopIteration
        idx = self.index
        self.index += 1
        return self._rcatvals[idx]


from xerparser.model.classes.rsrcrate import ResourceRate

class ResourceRates:

    _rsrcrates = []

    def __init__(self):
        self.index = 0

    def add(self, params):
        self._rsrcrates.append(ResourceRate(params))

    @classmethod
    def find_by_id(cls, id):
        obj = list(filter(lambda x: x.actv_code_type_id == id, cls._rsrcrates))
        if len(obj) > 0:
            return obj[0]
        return obj

    @property
    def count(self):
        return len(self._rsrcrates)

    def __len__(self):
        return len(ResourceRates._rsrcrates)

    def __iter__(self):
        return self

    def __next__(self):
        if self.index >= len(self._rsrcrates):
            raise StopIteration
        idx = self.index
        self.index += 1
        return self._rsrcrates[idx]


from xerparser.model.classes.acttype import ActType

class ActTypes:

    _activitytypes = []

    def __init__(self):
        self.index = 0

    def add(self, params):
        self._activitytypes.append(ActType(params))

    @classmethod
    def find_by_id(cls, id):
        obj = list(filter(lambda x: x.actv_code_type_id == id, cls._activitytypes))
        if len(obj) > 0:
            return obj[0]
        return obj


    @staticmethod
    def count():
        return len(ActTypes._activitytypes)

    def __len__(self):
        return len(ActTypes._activitytypes)

    def __iter__(self):
        return self

    def __next__(self):
        if self.index >= len(self._activitytypes):
            raise StopIteration
        idx = self.index
        self.index += 1
        return self._activitytypes[idx]


from xerparser.model.tasks import Tasks


class WBS:
    obj_list = []

    def __init__(self, params):
        self.wbs_id = int(params.get('wbs_id').strip()) if params.get('wbs_id') else None
        self.proj_id = int(params.get('proj_id').strip()) if params.get('proj_id') else None
        self.obs_id = params.get('obs_id').strip()
        self.seq_num = params.get('seq_num').strip()
        self.est_wt = params.get('est_wt')
        self.proj_node_flag = params.get('proj_node_flag').strip()
        self.sum_data_flag = params.get('sum_data_flag').strip()
        self.status_code = params.get('status_code').strip()
        self.wbs_short_name = params.get('wbs_short_name').strip()
        self.wbs_name = params.get('wbs_name').strip()
        self.phase_id = params.get('phase_id').strip()
        self.parent_wbs_id = int(params.get('parent_wbs_id')) if params.get('parent_wbs_id') else None
        self.ev_user_pct = params.get('ev_user_pct').strip()
        self.ev_etc_user_value = params.get('ev_etc_user_value').strip()
        self.orig_cost = params.get('orig_cost').strip()
        self.indep_remain_total_cost = params.get('indep_remain_total_cost').strip()
        self.ann_dscnt_rate_pct = params.get('ann_dscnt_rate_pct').strip()
        self.dscnt_period_type = params.get('dscnt_period_type').strip()
        self.indep_remain_work_qty = params.get('indep_remain_work_qty').strip()
        self.anticip_start_date = params.get('anticip_start_date').strip()
        self.anticip_end_date = params.get('anticip_end_date').strip()
        self.ev_compute_type = params.get('ev_compute_type').strip()
        self.ev_etc_compute_type = params.get('ev_etc_compute_type').strip()
        self.guid = params.get('guid').strip()
        self.tmpl_guid = params.get('tmpl_guid').strip()
        self.plan_open_state = params.get('plan_open_state').strip() if params.get('plan_open_state') else None

        WBS.obj_list.append(self)

    def get_id(self):
        return self.wbs_id

    @classmethod
    def get_json(cls):
        root_nodes = list(filter(lambda x: WBS.find_by_id(x.parent_wbs_id) is None, cls.obj_list))
        print(root_nodes)
        json = dict()
        for node in root_nodes:
            json["node"] = node
            json["level"] = 0
            json["childs"] = []
            json["childs"].append(cls.get_childs(node, 0))
        print(json)
        return json

    @classmethod
    def get_childs(cls, node, level):
        nodes_lst = list(filter(lambda x: x.parent_wbs_id == node.wbs_id, cls.obj_list))
        nod = dict()
        for node in nodes_lst:
            nod["node"] = node
            nod["level"] = level + 1
            children = cls.get_childs(node, level + 1)
            nod["childs"] = []
            nod["childs"].append(children)
        return nod
    @classmethod
    def find_by_id(cls, ID):
        obj = list(filter(lambda x: x.wbs_id == ID, cls.obj_list))
        if obj:
            return obj[0]
        return None

    @staticmethod
    def find_by_project_id(project_id, wbs):
        return {k: v for k, v in wbs.items() if v.proj_id == project_id}

    @property
    def activities(self):
        return Tasks.activities_by_wbs_id(self.wbs_id)

    def __repr__(self):
        return self.wbs_name



class ResourceRate:
    obj_list = []

    def __init__(self, params):
        self.rsrc_rate_id = params.get('rsrc_rate_id').strip() if params.get('rsrc_rate_id') else None
        self.rsrc_id = params.get('rsrc_id').strip() if params.get('rsrc_id') else None
        self.max_qty_per_hr = params.get('max_qty_per_hr').strip() if params.get('max_qty_per_hr') else None
        self.cost_per_qty = params.get('cost_per_qty').strip() if params.get('cost_per_qty') else None
        self.start_date = params.get('start_date').strip() if params.get('start_date') else None
        self.shift_period_id = params.get('shift_period_id').strip() if params.get('shift_period_id') else None
        self.cost_per_qty2 = params.get('cost_per_qty2').strip() if params.get('cost_per_qty2') else None
        self.cost_per_qty3 = params.get('cost_per_qty3').strip() if params.get('cost_per_qty3') else None
        self.cost_per_qty4 = params.get('cost_per_qty4').strip() if params.get('cost_per_qty4') else None
        self.cost_per_qty5 = params.get('cost_per_qty5').strip() if params.get('cost_per_qty5') else None
        ResourceRate.obj_list.append(self)

    def get_id(self):
        return self.rsrc_rate_id

    @classmethod
    def find_by_id(cls, id):
        obj = list(filter(lambda x: x.rsrc_rate_id == id, cls.obj_list))
        if len(obj) > 0:
            obj = obj[0]
        else:
            None
        return obj

    @classmethod
    def find_by_resource_id(cls, id):
        obj = list(filter(lambda x: x.rsrc_rate_id == id, cls.obj_list))
        if len(obj) > 0:
            obj = obj[0]
        else:
            obj = None
        return obj

    def __repr__(self):
        return self.rsrc_id



from xerparser.model.classes.taskpred import TaskPred

class Predecessors:
    task_pred = []
    def __init__(self):
        self.index=0

    def find_by_id(self, code_id):
        obj = list(filter(lambda x: x.task_pred_id == code_id, self.task_pred))
        if len(obj) > 0:
            obj[0]
        else:
            obj = None
        return obj

    def add(self, params):
        pred = TaskPred(params)
        self.task_pred.append(pred)

    @property
    def relations(self):
        return self.task_pred

    @property
    def leads(self):
        return list(filter(lambda x: x.lag_hr_cnt < 0 if x.lag_hr_cnt else None, self.task_pred))

    @property
    def finish_to_start(self):
        return list(filter(lambda x: x.pred_type == 'PR_FS', self.task_pred))
    @staticmethod
    def get_successors(act_id):
        succ = list(filter(lambda x: x.pred_task_id == act_id, Predecessors.task_pred))
        return succ

    @staticmethod
    def get_predecessors(act_id):
        succ = list(filter(lambda x: x.task_id == act_id, Predecessors.task_pred))
        return succ

    @staticmethod
    def count():
        return len(Predecessors.task_pred)

    def __len__(self):
        return len(self.task_pred)

    def __iter__(self):
        return self

    def __next__(self):
        if self.index >= len(self.task_pred):
            raise StopIteration
        idx = self.index
        self.index += 1
        return self.task_pred[idx]




from xerparser.model.classes.taskactv import TaskActv


class TaskActvs:
    _taskactvs = []

    def __init__(self):
        self.index = 0

    def add(self, params):
        self._taskactvs.append(TaskActv(params))

    @classmethod
    def find_by_code_id(cls, id):
        obj = list(filter(lambda x: x.actv_code_id == id, cls._taskactvs))
        if len(obj) > 0:
            return obj
        return obj

    @classmethod
    def find_by_activity_id(cls, id):
        obj = list(filter(lambda x: x.task_id == id, cls._taskactvs))
        if len(obj) > 0:
            return obj
        return obj

    @staticmethod
    def count():
        return len(TaskActvs._taskactvs)

    def __len__(self):
        return len(self._taskactvs)

    def __iter__(self):
        return self

    def __next__(self):
        if self.index >= len(self._taskactvs):
            raise StopIteration
        idx = self.index
        self.index += 1
        return self._taskactvs[idx]




class UDFType:

    udf_type_id = None
    table_name = None
    udf_type_name = None
    udf_type_label = None
    logical_data_type = None
    super_flag = None
    indicator_expression = None
    summary_indicator_expression = None

    def __init__(self, params):

        self.udf_type_id = params.get('udf_type_id').strip() if params.get('udf_type_id') else None
        self.table_name = params.get('table_name').strip() if params.get('table_name') else None
        self.udf_type_name = params.get('udf_type_name').strip() if params.get('udf_type_name') else None
        self.udf_type_label = params.get('udf_type_label').strip() if params.get('udf_type_label') else None
        self.logical_data_type = params.get('logical_data_type').strip() if params.get('logical_data_type') else None
        self.super_flag = params.get('super_flag').strip() if params.get('super_flag') else None
        self.indicator_expression = params.get('indicator_expression').strip() if params.get('indicator_expression') else None
        self.summary_indicator_expression = params.get('summary_indicator_expression').strip() if params.get('summary_indicator_expression') else None

    def get_id(self):
        return self.udf_type_id

    def __repr__(self):
        return self.udf_type_name



from xerparser.model.acttypes import ActTypes

class ActivityCode:
    obj_list = []
    def __init__(self, params):
        # Unique ID generated by the system.
        self.actv_code_id = int(params.get('actv_code_id').strip()) if params.get('actv_code_id') else None
        # The parent Activity Code value in the user code value hierarchy.
        self.parent_actv_code_id = int(params.get('parent_actv_code_id')) if params.get('parent_actv_code_id') else None
        # The Activity Code type id acts as foreign key to activity code types table
        self.actv_code_type_id = int(params.get('actv_code_type_id').strip()) if params.get('actv_code_type_id') else None
        # The description of an Activity Code value.
        self.actv_code_name = params.get('actv_code_name').strip() if params.get('actv_code_name') else None
        # The value of the activity code.
        self.short_name = params.get('short_name').strip() if params.get('short_name') else None
        # Sequence number for sorting.
        self.seq_num = int(params.get('seq_num').strip()) if params.get('seq_num') else None
        self.color = params.get('color').strip() if params.get('color') else None
        self.total_assignments = int(float(params.get('total_assignments').strip())) if params.get('total_assignments') else None
        ActivityCode.obj_list.append(self)

    def get_id(self):
        return self.actv_code_id

    @classmethod
    def find_by_id(cls, id):
        obj = list(filter(lambda x: x.actv_code_id == id, ActivityCode.obj_list))
        if obj:
            return obj[0]
        return obj

    @classmethod
    def find_by_code(cls, code):
        """ This Function searches for activity code using ID code

        Args:
            code: ID code as defined in Primavera and not the database
            obj_list: list of activity codes that need to be searched

        Returns: an ActivityCode object that matches the supplied code

        """
        actv_code = list(filter(lambda x: x.short_name == code, ActivityCode.obj_list))[0]
        return actv_code

    @classmethod
    def get_parent(cls, id):
        obj = None
        c_actv_code = list(filter(lambda x: x.actv_code_id == id, ActivityCode.obj_list))[0]
        if c_actv_code:
            obj = cls.find_by_id(c_actv_code.actv_code_id, ActivityCode.obj_list)
        return obj

    @classmethod
    def get_children(cls, id, obj_list):
        childs = list(filter(lambda x: x.parent_actv_code_id == id, obj_list))
        return childs

    @property
    def type(self):
        return ActTypes.find_by_id(self.actv_code_type_id)

    def __repr__(self):
        return str(self.actv_code_id) + ' - ' + self.short_name + ' - ' + self.actv_code_name



from datetime import datetime

from xerparser.model.taskactvs import TaskActvs

from xerparser.model.predecessors import Predecessors
from xerparser.model.classes.calendar import Calendar
from xerparser.model.activitiyresources import ActivityResources

class Task:
    # obj_list = []

    def __init__(self, params):
        # Unique ID generated by the system.
        self.task_id = int(params.get('task_id')) if params.get('task_id') else None
        # project to which the activity belongs referenced by system generated unique id
        self.proj_id = int(params.get('proj_id')) if params.get('proj_id') else None
        # wbs element activity assigned to referenced by system unique id
        self.wbs_id = int(params.get('wbs_id')) if params.get('wbs_id') else None
        # calendar assigned to activity referenced by system unique id
        self.clndr_id = int(params.get('clndr_id')) if params.get('clndr_id') else None
        # The physical percent complete can either be user entered or calculated from the activity's weighted steps.
        #  There is a project setting specifying this.
        self.phys_complete_pct = float(params.get('phys_complete_pct')) if params.get('phys_complete_pct') else None
        # Indicates that the primary resource has sent feedback notes about this activity which have not been
        # reviewed yet.
        self.rev_fdbk_flag = bool(params.get('rev_fdbk_flag')) if params.get('rev_fdbk_flag') else None
        # The estimation weight for the activity, used for top-down estimation. Top-down estimation weights are used
        # to calculate the proportion of units that each activity receives in relation to the other activities within
        #  the same WBS. Top-down estimation distributes estimated units in a top-down manner to activities using the
        #  WBS hierarchy.

        self.est_wt = float(params.get('est_wt').strip()) if params.get('est_wt') else None
        # Indicates that the planned labor and nonlabor units for the activity will not be modified by top-down
        # estimation.
        self.lock_plan_flag = bool(params.get('lock_plan_flag')) if params.get('lock_plan_flag') else None
        # Identifies whether the actual and remaining cost for the expense are computed automatically using the
        # planned cost and the activity's schedule percent complete.  If this option is selected,
        # the actual/remaining cost are automatically updated when project actuals are applied.  This assumes the
        # expenses are made according to plan.
        self.auto_compute_act_flag = bool(params.get('auto_compute_act_flag')) if params.get('auto_compute_act_flag') else None
        # The activity percent complete type is one of ""Duration"", ""Units"", or ""Physical"". The percent complete
        #  type controls whether the Activity % Complete is tied to the Duration % Complete, the Units % Complete,
        # or the Physical % Complete for the activity. Set the percent complete type to ""Duration"" for activities
        # which are duration driven, for example, administration tasks and training classes.  Set the percent
        # complete type to ""Physical"" for activities which are work-product driven, for example, creating a
        # document or a product. Set the percent complete type to ""Units"" for activities which are work effort
        # driven, for example, providing a consulting service.
        self.complete_pct_type = params.get('complete_pct_type').strip() if params.get('complete_pct_type') else None
        # The type of activity, either  'Task Dependent', 'Resource Dependent', 'Level of Effort', 'Start Milestone'
        # or 'Finish Milestone'.   A Task Dependent activity is scheduled using the activity's calendar rather than
        # the calendars of the assigned resources.  A Resource Dependent activity is scheduled using the calendars of
        #  the assigned resources.  This type is used when several resources are assigned to the activity,
        # but they may work separately.  A Start/Finish Milestone is a zero-duration activity, marking a significant
        # start/end of project event. A Level of Effort activity has a duration which is determined by its dependent
        # activities. Administration-type activities are typically level of effort.
        self.task_type = params.get('task_type').strip() if params.get('task_type') else None
        # The duration type of the activity. One of ""Fixed Units per Time"", ""Fixed Duration"", or ""Fixed Units"".
        #   For Fixed Units per Time activities, the resource units per time are constant when the activity duration
        # or units are changed.  This type is used when an activity has fixed resources with fixed productivity
        # output per time period.  For Fixed Duration activities, the activity duration is constant as the units or
        # resource units per time are changed. This type is used when the activity is to be completed within a fixed
        # time period regardless of the resources assigned.  For Fixed Units activities, the activity units are
        # constant when the duration or resource units per time are changed. This type is used when the total amount
        # of work is fixed, and increasing the resources can decrease the activity duration.
        self.duration_type = params.get('duration_type').strip() if params.get('duration_type') else None
        # The current status of the activity, either Not Started, In Progress, or Completed.
        self.status_code = params.get('status_code').strip() if params.get('status_code') else None
        # A short ID which uniquely identifies the activity within the project.
        self.task_code = params.get('task_code').strip() if params.get('task_code') else None
        # The name of the activity. The activity name does not have to be unique.
        self.task_name = params.get('task_name').strip() if params.get('task_name') else None
        # Resource ID Name
        self.rsrc_id = int(params.get('rsrc_id').strip()) if params.get('rsrc_id') else None
        # The amount of time the wbs can be delayed before delaying the project finish date. Total float can be
        # computed as Late Start - Early Start or as Late Finish - Early Finish; this option can be set when running
        # the project scheduler.
        self.total_float_hr_cnt = float(params.get('total_float_hr_cnt').strip()) if params.get('total_float_hr_cnt') else None
        # The amount of time the activity can be delayed before delaying the start date of any successor activity.
        self.free_float_hr_cnt = float(params.get('free_float_hr_cnt')) if params.get('free_float_hr_cnt') else None
        # Remaining duration is the total working time from the activity remaining start date to the remaining finish
        #  date. The remaining working time is computed using the activity's calendar. Before the activity is
        # started, the remaining duration is the same as the Original Duration. After the activity is completed the
        # remaining duration is zero.
        self.remain_drtn_hr_cnt = float(params.get('remain_drtn_hr_cnt').strip()) if params.get('remain_drtn_hr_cnt') else 0
        # The total actual labor units for all child activities
        self.act_work_qty = float(params.get('act_work_qty')) if params.get('act_work_qty') else None
        # The remaining units for all labor resources assigned to the activity. The remaining units reflects the work
        #  remaining to be done for the activity. Before the activity is started, the remaining units are the same as
        #  the planned units. After the activity is completed, the remaining units are zero.
        self.remain_work_qty = float(params.get('remain_work_qty')) if params.get('remain_work_qty') else None
        # The planned units for all labor resources assigned to the activity.
        self.target_work_qty = float(params.get('target_work_qty')) if params.get('target_work_qty') else None
        # Original Duration is the planned working time for the resource assignment on the activity,
        # from the resource's planned start date to the planned finish date. The planned working time is computed
        # using the calendar determined by the Activity Type. Resource Dependent activities use the resource's
        # calendar; other activity types use the activity's calendar. This is the duration that Timesheets users
        # follow and the schedule variance is measured against.
        self.target_drtn_hr_cnt = float(params.get('target_drtn_hr_cnt').strip()) if params.get('target_drtn_hr_cnt') else None
        # The planned units for all nonlabor resources assigned to the activity.
        self.target_equip_qty = float(params.get('target_equip_qty')) if params.get('target_equip_qty') else None
        # The actual units for all nonlabor resources assigned to the activities under the WBS.
        self.act_equip_qty = float(params.get('act_equip_qty')) if params.get('act_equip_qty') else None
        # The remaining units for all nonlabor resources assigned to the activity. The remaining units reflects the
        # work remaining to be done for the activity.  Before the activity is started, the remaining units are the
        # same as the planned units. After the activity is completed, the remaining units are zero.
        self.remain_equip_qty = float(params.get('remain_equip_qty')) if params.get('remain_equip_qty') else None
        # The constraint date for the activity, if the activity has a constraint. The activity's constraint type
        # determines whether this is a start date or finish date.  Activity constraints are used by the project
        # scheduler.
        self.cstr_date = datetime.strptime(params.get('cstr_date'), '%Y-%m-%d %H:%M') if params.get('cstr_date') else None
        # The date on which the activity is actually started.
        self.act_start_date = datetime.strptime(params.get('act_start_date'), '%Y-%m-%d %H:%M') if params.get('act_start_date') else None
        # The date on which the activity is actually finished.
        self.act_end_date = datetime.strptime(params.get('act_end_date'), '%Y-%m-%d %H:%M') if params.get('act_end_date') else None
        # the activity late start date
        self.late_start_date = datetime.strptime(params.get('late_start_date'), '%Y-%m-%d %H:%M') if params.get('late_start_date') else None
        # The latest possible date the activity must finish without delaying the project finish date. This date is
        # computed by the project scheduler based on network logic, schedule constraints, and resource availability.
        self.late_end_date = datetime.strptime(params.get('late_end_date'), '%Y-%m-%d %H:%M') if params.get('late_end_date') else None
        # The date the activity is expected to be finished according to the progress made on the activity's work
        # products. The expected finish date is entered manually by people familiar with progress of the activity's
        # work products.
        self.expect_end_date = datetime.strptime(params.get('expect_end_date'), '%Y-%m-%d %H:%M') if params.get('expect_end_date') else None
        # The earliest possible date the remaining work for the activity can begin. This date is computed by the
        # project scheduler based on network logic, schedule constraints, and resource availability.
        self.early_start_date = datetime.strptime(params.get('early_start_date'), '%Y-%m-%d %H:%M') if params.get('early_start_date') else None
        # The earliest possible date the activity can finish. This date is computed by the project scheduler based on
        #  network logic, schedule constraints, and resource availability.
        self.early_end_date = datetime.strptime(params.get('early_end_date'), '%Y-%m-%d %H:%M') if params.get('early_end_date') else None
        # The date the remaining work for the activity is scheduled to begin. This date is computed by the project
        # scheduler but can be updated manually by the project manager.  Before the activity is started,
        # the remaining start date is the same as the planned start date.  This is the start date that Timesheets
        # users follow.
        self.restart_date = datetime.strptime(params.get('restart_date'), '%Y-%m-%d %H:%M') if params.get('restart_date') else None
        # The date the remaining work for the activity is scheduled to finish. This date is computed by the project
        # scheduler but can be updated manually by the project manager. Before the activity is started, the remaining
        #  finish date is the same as the planned finish date.  This is the finish date that Timesheets users follow.
        self.reend_date = datetime.strptime(params.get('reend_date'), '%Y-%m-%d %H:%M') if params.get('reend_date') else None
        # The date the activity is scheduled to begin. This date is computed by the project scheduler but can be
        # updated manually by the project manager. This date is not changed by the project scheduler after the
        # activity has been started.
        self.target_start_date = datetime.strptime(params.get('target_start_date'), '%Y-%m-%d %H:%M') if params.get('target_start_date') else None
        # The date the activity is scheduled to finish. This date is computed by the project scheduler but can be
        # updated manually by the project manager.  This date is not changed by the project scheduler after the
        # activity has been started.
        self.target_end_date = datetime.strptime(params.get('target_end_date'), '%Y-%m-%d %H:%M') if params.get('target_end_date') else None
        # Remaining late start date is calculated by the scheduler.
        self.rem_late_start_date = datetime.strptime(params.get('rem_late_start_date'), '%Y-%m-%d %H:%M') if params.get('rem_late_start_date') else None
        # Remaining late end date is calculated by the scheduler.
        self.rem_late_end_date = datetime.strptime(params.get('rem_late_end_date'), '%Y-%m-%d %H:%M') if params.get('rem_late_end_date') else None
        # The type of constraint applied to the activity start or finish date. Activity constraints are used by the
        # project scheduler.  Start date constraints are 'Start On', 'Start On or Before', 'Start On or After' and
        # 'Mandatory Start'.  Finish date constraints are 'Finish On', 'Finish On or Before', 'Finish On or After'
        # and 'Mandatory Finish'.  Another type of constraint, 'As Late as Possible', schedules the activity as late
        # as possible based on the available free float.
        self.cstr_type = params.get('cstr_type').strip() if params.get('cstr_type') else None
        self.priority_type = params.get('priority_type').strip() if params.get('priority_type') else None
        # The date progress is suspended on an activity.
        self.suspend_date = datetime.strptime(params.get('suspend_date').strip(), '%Y-%m-%d %H:%M') if params.get('suspend_date') else None
        # The date progress is resumed on an activity.
        self.resume_date = datetime.strptime(params.get('resume_date').strip(), '%Y-%m-%d %H:%M') if params.get('resume_date') else None
        self.float_path = params.get('float_path').strip() if params.get('float_path') else None
        # This field is computed by the project scheduler and identifies the order in which the activities were
        # processed within the float path.
        self.float_path_order = params.get('float_path_order').strip() if params.get('float_path_order') else None
        self.guid = params.get('guid').strip() if params.get('guid') else None
        self.tmpl_guid = params.get('tmpl_guid').strip() if params.get('tmpl_guid') else None
        # The second constraint date for the activity, if the activity has a constraint.
        self.cstr_date2 = datetime.strptime(params.get('cstr_date2'), '%Y-%m-%d %H:%M') if params.get('cstr_date2') else None
        # The second type of constraint applied to the activity start or finish date.
        self.cstr_type2 = params.get('cstr_type2').strip() if params.get('cstr_type2') else None
        self.driving_path_flag = bool(params.get('driving_path_flag')) if params.get('driving_path_flag') else None
        # The actual this period units for all labor resources assigned to the activity.
        self.act_this_per_work_qty = float(params.get('act_this_per_work_qty')) if params.get('act_this_per_work_qty') else None
        # The actual this period units for all nonlabor resources assigned to the activity.
        self.act_this_per_equip_qty = float(params.get('act_this_per_equip_qty')) if params.get('act_this_per_equip_qty') else None
        # The External Early Start date is the date the external relationship was scheduled to finish.  This date may
        #  be used to calculate the start date of the current activity during scheduling.  This field is populated on
        #  import when an external relationship is lost.
        try:
            self.external_early_start_date = datetime.strptime(params.get('external_early_start_date').strip(), '%Y-%m-%d %H:%M') if params.get('external_early_start_date') else None
            self.external_late_end_date = datetime.strptime(params.get('external_late_end_date'), '%Y-%m-%d %H:%M') if params.get('external_late_end_date') else None
        except:
            pass
        self.create_date = params.get('create_date').strip() if params.get('create_date') else None
        self.update_date = params.get('update_date').strip() if params.get('update_date') else None
        self.create_user = params.get('create_user').strip() if params.get('create_user') else None
        self.update_user = params.get('update_user').strip() if params.get('update_user') else None
        self.location_id = params.get('location_id').strip() if params.get('location_id') else None
        self.calendar = Calendar.find_by_id(self.clndr_id)
        # self.wbs = WBS.find_by_id(int(self.wbs_id) if self.wbs_id else None)
        # Task.obj_list.append(self)

    @property
    def id(self):
        return self.task_id

    @property
    def totalfloat(self):
        if self.total_float_hr_cnt:
            tf = float(self.total_float_hr_cnt)/8.0
        else:
            return None
        return tf

    @property
    def resources(self):
        return ActivityResources.find_by_activity_id(self.task_id)

    @property
    def activitycodes(self):
        return TaskActvs.find_by_activity_id(self.task_id)


    @property
    def duration(self):
        dur = None
        if self.target_drtn_hr_cnt:
            dur = float(self.target_drtn_hr_cnt) / float(self.calendar.day_hr_cnt)
        else:
            dur =0.0
        return dur

    @property
    def constraints(self):
        if self.cstr_type == None or self.cstr_date == None:
            return  None
        return {"ConstraintType": self.cstr_type,
                "ConstrintDate": self.cstr_date
               }

    @property
    def start_date(self):
        if self.act_start_date:
            return self.act_start_date
        else:
            return self.target_start_date

    @property
    def end_date(self):
        if self.act_end_date:
            return self.act_end_date
        else:
            return self.target_end_date

    @property
    def successors(self):
        suss = Predecessors.get_successors(self.task_id)
        return suss

    @property
    def predecessors(self):
        return Predecessors.get_predecessors(self.task_id)

    def __repr__(self):
        return self.task_code







class ResourceCurve:
    obj_list = []

    def __init__(self, params):
        self.curv_id = int(params.get('curv_id')) if params.get('curv_id') else None
        self.curv_name = params.get('curv_id').strip() if params.get('curv_id') else None
        self.default_flag = params.get('default_flag') if params.get('default_flag') else None
        self.pct_usage_0 = float(params.get('pct_usage_0')) if params.get('pct_usage_0') else None
        self.pct_usage_1 = float(params.get('pct_usage_1')) if params.get('pct_usage_1') else None
        self.pct_usage_2 = float(params.get('pct_usage_2')) if params.get('pct_usage_2') else None
        self.pct_usage_3 = float(params.get('pct_usage_3')) if params.get('pct_usage_3') else None
        self.pct_usage_4 = float(params.get('pct_usage_4')) if params.get('pct_usage_4') else None
        self.pct_usage_5 = float(params.get('pct_usage_5')) if params.get('pct_usage_5') else None
        self.pct_usage_6 = float(params.get('pct_usage_6')) if params.get('pct_usage_6') else None
        self.pct_usage_7 = float(params.get('pct_usage_7')) if params.get('pct_usage_7') else None
        self.pct_usage_8 = float(params.get('pct_usage_8')) if params.get('pct_usage_8') else None
        self.pct_usage_9 = float(params.get('pct_usage_9')) if params.get('pct_usage_9') else None
        self.pct_usage_10 = float(params.get('pct_usage_10')) if params.get('pct_usage_10') else None
        self.pct_usage_11 = float(params.get('pct_usage_11')) if params.get('pct_usage_11') else None
        self.pct_usage_12 = float(params.get('pct_usage_12')) if params.get('pct_usage_12') else None
        self.pct_usage_13 = float(params.get('pct_usage_13')) if params.get('pct_usage_13') else None
        self.pct_usage_14 = float(params.get('pct_usage_14')) if params.get('pct_usage_14') else None
        self.pct_usage_15 = float(params.get('pct_usage_15')) if params.get('pct_usage_15') else None
        self.pct_usage_16 = float(params.get('pct_usage_16')) if params.get('pct_usage_16') else None
        self.pct_usage_17 = float(params.get('pct_usage_17')) if params.get('pct_usage_17') else None
        self.pct_usage_18 = float(params.get('pct_usage_18')) if params.get('pct_usage_18') else None
        self.pct_usage_19 = float(params.get('pct_usage_19')) if params.get('pct_usage_19') else None
        self.pct_usage_20 = float(params.get('pct_usage_20')) if params.get('pct_usage_20') else None

        ResourceCurve.obj_list.append(self)

    @classmethod
    def find_by_id(cls, id):
        obj = list(filter(lambda x: x.curv_id == id, cls.obj_list))[0]
        return obj

    def __repr__(self):
        return self.curv_name



from xerparser.model.classes.activitycode import ActivityCode


class ActivityCodes:

    _activitycodes = []

    def __init__(self):
        self.index = 0

    def add(self, params):
        self._activitycodes.append(ActivityCode(params))

    @staticmethod
    def count():
        return len(ActivityCodes._activitycodes)

    def __len__(self):
        return len(ActivityCodes._activitycodes)

    def __iter__(self):
        return self

    def __next__(self):
        if self.index >= len(self._activitycodes):
            raise StopIteration
        idx = self.index
        self.index += 1
        return self._activitycodes[idx]




class OBS:
    obj_list = []

    def __init__(self, params):
        # Unique ID generated by the system.
        self.obs_id = int(params.get('obs_id').strip()) if params.get('obs_id') else None
        # The parent OBS value in the OBS hierarchy.
        self.parent_obs_id = int(params.get('parent_obs_id').strip()) if params.get('parent_obs_id') else None
        # Global Unique Identifier generated by the system.
        self.guid = params.get('guid').strip() if params.get('guid') else None
        # Sequence number used for sorting.
        self.seq_num = int(params.get('seq_num').strip()) if params.get('seq_num') else None
        # The name of the person/role in the organization.
        self.obs_name = params.get('obs_name').strip() if params.get('obs_name') else None
        # The description of the person/role in the organization.
        self.obs_descr = params.get('obs_descr').strip() if params.get('obs_descr') else None

        OBS.obj_list.append(self)

    @classmethod
    def get_json(cls):
        root_nodes = list(filter(lambda x: x.parent_obs_id is None, cls.obj_list))
        # print(root_nodes)
        json = dict()
        for node in root_nodes:
            json["node"] = node
            json["level"] = 0
            json["childs"] = cls.get_childs(node, 0)
        print(json)

    @classmethod
    def get_childs(cls, node, level):
        nodes_lst = list(filter(lambda x: x.parent_obs_id == node.obs_id, cls.obj_list))
        nod = dict()
        for node in nodes_lst:
            nod["node"] = node
            nod["level"] = level + 1
            children = cls.get_childs(node, level+1)
            nod["childs"] = children if children else []
        return nod

    def get_id(self):
        return self.obs_id

    @classmethod
    def find_by_id(cls, id):
        obj = list(filter(lambda x: x.obs_id == id, cls.obj_list))
        if obj:
            return obj[0]
        return obj

    def __repr__(self):
        return self.obs_name



from xerparser.model.classes.task import Task
from xerparser.model.classes.taskpred import TaskPred
from xerparser.model.classes.taskactv import TaskActv
from xerparser.model.predecessors import Predecessors

class Tasks:
    """
    This class is a collection of tasks that controls functionalities to search, add, update and delete tasks
    """
    _tasks = []

    def __init__(self):
        self.index = 0

    def add(self, params):
        task = Task(params)
        self._tasks.append(task)

    @property
    def activities(self):
        return self._tasks

    @property
    def count(self):
        return len(self._tasks)

    @property
    def has_no_successor(self):
        objs = list(filter(lambda x: x.task_id not in [z.pred_task_id for z in TaskPred.obj_list], self._tasks))
        return objs

    @property
    def has_no_predecessor(self):
        objs = list(filter(lambda x: x.task_id not in [z.task_id for z in TaskPred.obj_list], self._tasks))
        return objs

    def __len__(self):
        return len(self._tasks)

    def __repr__(self):
        return [x.task_code for x in self._tasks]

    def __str__(self):
        return str([str(x.task_code) for x in self._tasks])

    @property
    def constraints(self):
        lst = [x.constraints if x.constraints is not None else None for x in self._tasks]
        # print(lst)
        return list(filter(lambda x: x is not None, lst))
    @classmethod
    def find_by_id(cls, id):
        obj = list(filter(lambda x: x.task_id == id, cls._tasks))
        if len(obj) > 0:
            return obj[0]
        return obj

    @classmethod
    def find_by_code(cls, code):
        obj = list(filter(lambda x: x.task_code == code, cls._tasks))
        if len(obj) > 0:
            return obj[0]
        return obj

    def duration_greater_than(self, duration):
        obj = list(filter(lambda x: x.target_drtn_hr_cnt > duration * float(self.calendar.day_hr_cnt), self._tasks))
        if obj:
            return obj
        return obj

    def float_less_than(self, Tfloat):
        objs = list(filter(lambda x: x.status_code != "TK_Complete", self._tasks))
        obj = list(filter(lambda x: x.total_float_hr_cnt < Tfloat * float(x.calendar.day_hr_cnt), objs))
        if obj:
            return obj
        return obj

    def float_greater_than(self, Tfloat):
        objs = list(filter(lambda x: x.status_code != "TK_Complete", self._tasks))
        obj = list(filter(lambda x: x.total_float_hr_cnt > Tfloat * float(x.calendar.day_hr_cnt), objs))
        if obj:
            return obj
        return obj

    def float_within_range(self, float1, float2):
        obj = None
        objs = list(filter(lambda x: x.status_code != "TK_Complete", self._tasks))
        if float1 < float2:
            obj = list(filter(lambda x: x.total_float_hr_cnt >= float1 * float(x.calendar.day_hr_cnt) and x.total_float_hr_cnt <= float2 * float(x.calendar.day_hr_cnt), objs))
            if obj:
                return obj
        return obj

    def float_within_range_exclusive(self, float1, float2):
        obj = None
        objs = list(filter(lambda x: x.status_code != "TK_Complete", self._tasks))
        if float1 < float2:
            obj = list(filter(lambda x: x.total_float_hr_cnt > float1 * float(x.calendar.day_hr_cnt) and x.total_float_hr_cnt < float2 * float(x.calendar.day_hr_cnt), objs))
            if obj:
                return obj
        return obj

    def activities_by_status(self, status):
        objs = list(filter(lambda x: x.status_code == status, self._tasks))
        return objs

    @staticmethod
    def activities_by_wbs_id(id):
        objs = list(filter(lambda x: x.wbs_id == id, Tasks._tasks))
        return objs


    def activities_by_activity_code_id(self, id):
        objs = list(filter(lambda x: x.actv_code_id == id, TaskActv.obj_list))
        activities = []
        for obj in objs:
            activities.append(self.find_by_id(obj.task_id))
        return activities

    
    def no_predecessors(self):
        objs = list(filter(lambda x: x.task_id not in [z.task_id for z in TaskPred.obj_list], self._tasks))
        return objs

    def no_successors(self):
        objs = list(filter(lambda x: x.task_id not in [z.pred_task_id for z in TaskPred.obj_list], self._tasks))
        return objs

    def activities_with_hard_contratints(self):
        obj = list(filter(lambda x: x.cstr_type == "CS_MEO" or x.cstr_type == "CS_MSO", self._tasks))
        return obj

    def activities_by_type(self, type):
        obj = list(filter(lambda x: x.cstr_type == type, self._tasks))
        return obj

    @staticmethod
    def get_by_project(id):
        obj = list(filter(lambda x: x.proj_id == id, Tasks._tasks))
        return obj

    def __iter__(self):
        return self

    def __next__(self):
        if self.index >= len(self._tasks):
            raise StopIteration
        idx = self.index
        self.index +=1
        return self._tasks[idx]



from xerparser.model.classes.calendar_data import CalendarData


class Calendar:

    obj_list = []
    working_days = dict()
    exceptions = []

    def __init__(self, params):
        # Unique ID generated by the system.
        self.clndr_id = int(params.get('clndr_id').strip()) if params.get('clndr_id') else None
        # Identifies the default global calendar (applies to global calendars only).
        self.default_flag = params.get('default_flag').strip() if params.get('default_flag') else None
        # The name of the calendar.
        self.clndr_name = params.get('clndr_name').strip() if params.get('clndr_name') else None
        self.proj_id = params.get('proj_id').strip() if params.get('proj_id') else None
        # The global calendar to which this calendar is linked.  Any changes to the global calendar are automatically
        #  propagated to this calendar.
        self.base_clndr_id = params.get('base_clndr_id').strip() if params.get('base_clndr_id') else None
        # Date of last changes to calendar.
        self.last_chng_date = params.get('last_chng_date').strip() if params.get('last_chng_date') else None
        # The calendar type - either global, resource, or project. Global calendars can be assigned to projects and
        # resources. Resource calendars can be assigned only to resources. Project calendars are specific to projects.
        self.clndr_type = params.get('clndr_type').strip() if params.get('clndr_type') else None
        # The number of work hours per day. This conversion factor is used for displaying time units and durations in
        #  the user's selected display formats.
        self.day_hr_cnt = float(params.get('day_hr_cnt')) if params.get('day_hr_cnt') else None
        # The number of work hours per week. This conversion factor is used for displaying time units and durations
        # in the user's selected display formats.
        self.week_hr_cnt = float(params.get('week_hr_cnt')) if params.get('week_hr_cnt') else None
        # The number of work hours per month. This conversion factor is used for displaying time units and durations
        # in the user's selected display formats.
        self.month_hr_cnt = float(params.get('month_hr_cnt')) if params.get('month_hr_cnt') else None
        # The number of work hours per year. This conversion factor is used for displaying time units and durations
        # in the user's selected display formats.
        self.year_hr_cnt = float(params.get('year_hr_cnt')) if params.get('year_hr_cnt') else None
        #
        self.rsrc_private = params.get('rsrc_private').strip() if params.get('rsrc_private') else None

        self.clndr_data = params.get('rsrc_private').strip() if params.get('rsrc_private') else None

        if self.clndr_data:
            c = CalendarData(self.clndr_data)
            self.working_days = c.get_days()
            self.exceptions = c.get_exceptions()
        Calendar.obj_list.append(self)

    def get_id(self):
        return self.clndr_id

    @classmethod
    def find_by_id(cls, id):
        obj = list(filter(lambda x: x.clndr_id == id, cls.obj_list))
        if obj:
            return obj[0]
        return None

    def __repr__(self):
        return self.clndr_name + ' : ' + str(self.day_hr_cnt) + '\n' + self.clndr_data


from setuptools import setup

def readme():
    with open('README.md') as f:
        return f.read()

setup(
    name='PyP6Xer',
    version='1.011.06',
    long_description=readme(),
    long_description_content_type='text/markdown',
    packages=['xerparser', 'xerparser.model', 'xerparser.model.classes'],
    url='',
    license='GNU GENERAL PUBLIC LICENSE',
    author='Hassan',
    author_email='hassan.emam@hotmail.com',
    install_requires=[
        'tqdm==4.50.0'
      ],
    description='Parser for XER written in Python'
)





class RoleRate:
    obj_list = []

    def __init__(self, params):
        self.role_rate_id = int(params.get('role_rate_id').strip()) if params.get('role_rate_id') else None
        self.role_id = int(params.get('role_id').strip()) if params.get('role_id') else None
        self.cost_per_qty = float(params.get('cost_per_qty').strip()) if params.get('cost_per_qty') else None
        self.cost_per_qty2 = float(params.get('cost_per_qty2').strip())if params.get('cost_per_qty2') else None
        self.cost_per_qty3 = float(params.get('cost_per_qty3').strip()) if params.get('cost_per_qty3') else None
        self.cost_per_qty4 = float(params.get('cost_per_qty4').strip()) if params.get('cost_per_qty4') else None
        self.cost_per_qty5 = float(params.get('cost_per_qty5').strip()) if params.get('cost_per_qty5') else None

        RoleRate.obj_list.append(self)

    @classmethod
    def find_by_id(cls, id):
        obj = list(filter(lambda x: x.role_rate_id == id, cls.obj_list))[0]
        return obj

    @classmethod
    def find_by_role_id(cls, id):
        obj = list(filter(lambda x: x.role_id == id, cls.obj_list))
        if len(obj) > 0:
            obj = obj[0]
        else:
            obj = None
        return obj

    def __repr__(self):
        return str(self.role_rate_id)



from xerparser.model.classes.taskrsrc import TaskRsrc
from xerparser.model.classes.rsrc import Resource

class ActivityResources:

    _taskrsrc = []

    def __init__(self):
        self.index = 0

    def add(self, params):
        self._taskrsrc.append(TaskRsrc(params))

    @classmethod
    def find_by_id(cls, id):
        obj = list(filter(lambda x: x.taskrsrc_id == id , cls._taskrsrc))
        if len(obj) > 0:
            return obj[0]
        return obj

    @classmethod
    def find_by_activity_id(cls, id):
        obj = list(filter(lambda x: x.task_id == id and x.rsrc_id, cls._taskrsrc))
        obj1 = [{Resource.find_by_id(x.rsrc_id): {"BL_QTY":x.target_qty, "ACT_QTY": x.act_reg_qty,\
                                                  "REM_QTY": x.remain_qty}} for x in obj]
        return obj

    @property
    def count(self):
        return len(self._taskrsrc)

    def __len__(self):
        return len(ActivityResources._taskrsrc)

    def __iter__(self):
        return self

    def __next__(self):
        if self.index >= len(self._taskrsrc):
            raise StopIteration
        idx = self.index
        self.index += 1
        return self._taskrsrc[idx]


from xerparser.model.classes.wbs import WBS

class WBSs:
    _wbss = []

    def __init__(self):
        self.index = 0

    def add(self, params):
        wbs = WBS(params)
        self._wbss.append(wbs)

    @staticmethod
    def get_by_project(id):
        return list(filter(lambda x: x.proj_id == id, WBSs._wbss))



    def __iter__(self):
        return self

    def __next__(self):
        if self.index >= len(self._wbss):
            raise StopIteration
        idx = self.index
        self.index +=1
        return self._wbss[idx]




class Account:
    obj_list = []

    def __init__(self, params):
        self.acct_id = int(params.get('acct_id').strip()) if params.get('acct_id') else None
        self.parent_acct_id = int(params.get('parent_acct_id').strip()) if params.get('parent_acct_id') else None
        self.acct_seq_num = int(params.get('acct_seq_num').strip()) if params.get('acct_seq_num') else None
        self.acct_name = params.get('acct_name').strip() if params.get('acct_name') else None
        self.acct_short_name = params.get('acct_short_name').strip() if params.get('acct_short_name') else None
        self.acct_descr = params.get('acct_descr').strip() if params.get('acct_descr') else None
        Account.obj_list.append(self)

    @classmethod
    def find_by_id(cls, id):
        obj = list(filter(lambda x: x.acct_id == id, cls.obj_list))[0]
        return obj

    def __repr__(self):
        return self.acct_name



from xerparser.model.resources import Resources
class TaskRsrc:
    taskrsrc_id = None
    task_id = None
    proj_id = None
    cost_qty_link_flag = None
    role_id = None
    acct_id = None
    rsrc_id = None
    pobs_id = None
    skill_level = None
    remain_qty = None
    target_qty = None
    remain_qty_per_hr = None
    target_lag_drtn_hr_cnt = None
    target_qty_per_hr = None
    act_ot_qty = None
    act_reg_qty = None
    relag_drtn_hr_cnt = None
    ot_factor = None
    cost_per_qty = None
    target_cost = None
    act_reg_cost = None
    act_ot_cost = None
    remain_cost = None
    act_start_date = None
    act_end_date = None
    restart_date = None
    reend_date = None
    target_start_date = None
    target_end_date = None
    rem_late_start_date = None
    rem_late_end_date = None
    rollup_dates_flag = None
    target_crv = None
    remain_crv = None
    actual_crv = None
    ts_pend_act_end_flag = None
    guid = None
    rate_type = None
    act_this_per_cost = None
    act_this_per_qty = None
    curv_id = None
    rsrc_type = None
    cost_per_qty_source_type = None
    create_user = None
    create_date = None
    has_rsrchours = None
    taskrsrc_sum_id = None

    def __init__(self, params):
        self.taskrsrc_id = params.get('taskrsrc_id').strip() if params.get('taskrsrc_id') else None
        self.task_id = int(params.get('task_id').strip()) if params.get('task_id') else None
        self.proj_id = params.get('proj_id').strip() if params.get('proj_id') else None
        self.cost_qty_link_flag = params.get('cost_qty_link_flag').strip() if params.get('cost_qty_link_flag') else None
        self.role_id = params.get('role_id').strip() if params.get('role_id') else None
        self.acct_id = params.get('acct_id').strip() if params.get('acct_id') else None
        self.rsrc_id = int(params.get('rsrc_id').strip()) if params.get('rsrc_id') else None
        self.pobs_id = params.get('pobs_id').strip() if params.get('pobs_id') else None
        self.skill_level = params.get('skill_level').strip() if params.get('skill_level') else None
        self.remain_qty = float(params.get('remain_qty').strip()) if params.get('remain_qty') else None
        self.target_qty = float(params.get('target_qty').strip()) if params.get('target_qty') else None
        self.remain_qty_per_hr = float(params.get('remain_qty_per_hr').strip()) if params.get('remain_qty_per_hr') else None
        self.target_lag_drtn_hr_cnt = float(params.get('target_lag_drtn_hr_cnt').strip()) if params.get('target_lag_drtn_hr_cnt') else None
        self.target_qty_per_hr = params.get('target_qty_per_hr').strip() if params.get('target_qty_per_hr') else None
        self.act_ot_qty = params.get('act_ot_qty').strip() if params.get('act_ot_qty') else None
        self.act_reg_qty = params.get('act_reg_qty').strip() if params.get('act_reg_qty') else None
        self.relag_drtn_hr_cnt = params.get('relag_drtn_hr_cnt').strip() if params.get('relag_drtn_hr_cnt') else None
        self.ot_factor = params.get('ot_factor').strip() if params.get('ot_factor') else None
        self.cost_per_qty = params.get('cost_per_qty').strip() if params.get('cost_per_qty') else None
        self.target_cost = params.get('target_cost').strip() if params.get('target_cost') else None
        self.act_reg_cost = float(params.get('act_reg_cost').strip()) if params.get('act_reg_cost') else None
        self.act_ot_cost = params.get('act_ot_cost').strip() if params.get('act_ot_cost') else None
        self.remain_cost = params.get('remain_cost').strip() if params.get('remain_cost') else None
        self.act_start_date = params.get('act_start_date').strip() if params.get('act_start_date') else None
        self.act_end_date = params.get('act_end_date').strip() if params.get('act_end_date') else None
        self.restart_date = params.get('restart_date').strip() if params.get('restart_date') else None
        self.reend_date = params.get('reend_date').strip() if params.get('reend_date') else None
        self.target_start_date = params.get('target_start_date').strip() if params.get('target_start_date') else None
        self.target_end_date = params.get('target_end_date').strip() if params.get('target_end_date') else None
        self.rem_late_start_date = params.get('rem_late_start_date').strip() if params.get('rem_late_start_date') else None
        self.rem_late_end_date = params.get('rem_late_end_date').strip() if params.get('rem_late_end_date') else None
        self.rollup_dates_flag = params.get('rollup_dates_flag').strip() if params.get('rollup_dates_flag') else None
        self.target_crv = params.get('target_crv').strip() if params.get('target_crv') else None
        self.remain_crv = params.get('remain_crv').strip() if params.get('remain_crv') else None
        self.actual_crv = params.get('actual_crv').strip() if params.get('actual_crv') else None
        self.ts_pend_act_end_flag = params.get('ts_pend_act_end_flag').strip() if params.get('ts_pend_act_end_flag') else None
        self.guid = params.get('guid').strip() if params.get('guid') else None
        self.rate_type = params.get('rate_type').strip() if params.get('rate_type') else None
        self.act_this_per_cost = params.get('act_this_per_cost').strip() if params.get('act_this_per_cost') else None
        self.act_this_per_qty = params.get('act_this_per_cost').strip() if params.get('act_this_per_cost') else None
        self.curv_id = params.get('curv_id').strip() if params.get('curv_id') else None
        self.rsrc_type = params.get('rsrc_type').strip() if params.get('rsrc_type') else None
        self.cost_per_qty_source_type = params.get('cost_per_qty_source_type').strip() if params.get('cost_per_qty_source_type') else None
        self.create_user = params.get('create_user').strip() if params.get('create_user') else None
        self.create_date = params.get('create_date').strip() if params.get('create_date') else None
        self.has_rsrchours = params.get('has_rsrchours').strip() if params.get('has_rsrchours') else None
        self.taskrsrc_sum_id = params.get('taskrsrc_sum_id').strip() if params.get('taskrsrc_sum_id') else None

    def get_id(self):
        return self.taskrsrc_id
    @property
    def resource(self):
        return Resources.get_resource_by_id(self.rsrc_id)

    @staticmethod
    def find_by_id(code_id, activity_code_dict):
        return {k: v for k, v in activity_code_dict.items() if v.actv_code_id == code_id}

    def __repr__(self):
        return str(self.task_id) + '->' + str(self.rsrc_id) + ' = ' #+ self.target_qty
